"""Transform DSL data to GSE data.

Notes:
    Works in a similar way to IDL spedas dsl2gse.pro
"""

import logging
import numpy as np
from copy import deepcopy
#import pytplot

import pyspedas
from pyspedas.cotrans.cotrans_lib import subgei2gse
from pytplot import data_exists, del_data, store_data, get_data, set_coords, get_coords
from pyspedas.themis import autoload_support


def dsl2gse(name_in: str, name_out: str, isgsetodsl: bool = False, ignore_input_coord: bool = False,
            probe: str=None, use_spinaxis_corrections: bool=True) -> int:
    """Transform dsl to gse.

    Parameters
    ----------
        name_in: str
            Name of input pytplot variable (e.g. 'tha_fgl_dsl')
        name_out: str
            Name of output pytplot variable (e.g. 'tha_fgl_gse')
        isgsetodsl: bool
            If False (default) then DSL to GSE.
            If True, then GSE to DSL.
        ignore_input_coord: bool
            if False (default), do not check the input coordinate system
            if True, fail and return 0 if input coordinate does not match the requested transform.
        probe: str
            Usually optional, if the probe can be inferred from the input variable name (e.g. tha_xxx_yyy).
            Otherwise, one of ['a','b','c','d','e','f']
        use_spinaxis_corrections: bool
            If True (default), use spin axis corrections from V03 state CDFs when available.
            If False, use the uncorrected spin axis variables.

    Returns
    -------
        1 for successful completion.

    """
    needed_vars = [name_in]
    c = [value for value in needed_vars if data_exists(value)]
    if len(c) < 1:
        logging.error("Variables needed: " + str(needed_vars))
        m = [value for value in needed_vars if value not in c]
        logging.error("Variables missing: " + str(m))
        logging.error("Please load missing variables.")
        return 0

    if probe is None:
        probe = name_in[2]

    if not ignore_input_coord:
        in_coord = get_coords(name_in)
        if in_coord is None:
            in_coord = "None"
        if isgsetodsl and (in_coord.lower() != 'gse'):
            logging.error("GSE to DSL transform requested, but input coordinate system is " + in_coord)
            return 0
        if not isgsetodsl and (in_coord.lower() != 'dsl'):
            logging.error("DSL to GSE transform requested, but input coordinate system is " + in_coord)
            return 0

    autoload_support(varname=name_in, probe=probe, spinaxis=True)

    if use_spinaxis_corrections:
        spinras = 'th' + probe + '_spinras_corrected'
        spindec = 'th' + probe + '_spindec_corrected'
    else:
        spinras = 'th' + probe + '_spinras'
        spindec = 'th' + probe + '_spindec'

    # Interpolate spinras and spindec
    spinnames_in = [spinras, spindec]
    hiras_name = spinras + '_hires'
    hidec_name = spindec + '_hires'
    hi_names = [hiras_name, hidec_name]

    # If new names exist, delete the variables
    if data_exists(hiras_name):
        del_data(hiras_name)
    if data_exists(hidec_name):
        del_data(hidec_name)

    pyspedas.tinterpol(spinnames_in, name_in, method="linear",
                       newname=hi_names, suffix='')

    # Get data
    data_in = get_data(name_in)
    meta_in = get_data(name_in, metadata=True)
    meta_copy = deepcopy(meta_in)
    data_ras = get_data(hiras_name)
    data_dec = get_data(hidec_name)

    # Make a unit vector that points along the spin axis
    spla = (90.0 - (data_dec[1])) * np.pi / 180.0
    splo = data_ras[1] * np.pi / 180.0
    # spherical to cartesian
    zscs0 = np.sin(spla) * np.cos(splo)
    zscs1 = np.sin(spla) * np.sin(splo)
    zscs2 = np.cos(spla)
    znorm = np.sqrt(zscs0 * zscs0 + zscs1 * zscs1 + zscs2 * zscs2)
    zscs0 = np.divide(zscs0, znorm)
    zscs1 = np.divide(zscs1, znorm)
    zscs2 = np.divide(zscs2, znorm)
    zscs = np.column_stack((zscs0, zscs1, zscs2))

    # unit vector that points along the spin axis in GSE
    trgse = subgei2gse(data_in[0], zscs)
    zgse = trgse
    sun = [1.0, 0.0, 0.0]
    my_y = np.cross(zgse, sun)
    ynorm = np.sqrt(my_y[:, 0] * my_y[:, 0] + my_y[:, 1] * my_y[:, 1] + my_y[:, 2] * my_y[:, 2])
    my_y[:, 0] = np.divide(my_y[:, 0], ynorm)
    my_y[:, 1] = np.divide(my_y[:, 1], ynorm)
    my_y[:, 2] = np.divide(my_y[:, 2], ynorm)
    my_x = np.cross(my_y, zgse)
    xnorm = np.sqrt(my_x[:, 0] * my_x[:, 0] + my_x[:, 1] * my_x[:, 1] + my_x[:, 2] * my_x[:, 2])
    my_x[:, 0] = np.divide(my_x[:, 0], xnorm)
    my_x[:, 1] = np.divide(my_x[:, 1], xnorm)
    my_x[:, 2] = np.divide(my_x[:, 2], xnorm)

    yscs = np.column_stack((zgse[:, 1] * sun[2] - zgse[:, 2] * sun[1],
                            zgse[:, 2] * sun[0] - zgse[:, 0] * sun[2],
                            zgse[:, 0] * sun[1] - zgse[:, 1] * sun[0]))
    # yscs_norm = np.sqrt(yscs[:,0] ** 2.0 + yscs[:,1] ** 2.0 + yscs[:,2] ** 2.0)
    # yscs = np.divide(yscs, yscs_norm)
    xscs = np.column_stack((yscs[:, 1] * zgse[:, 2] - yscs[:, 2] * zgse[:, 1],
                            yscs[:, 2] * zgse[:, 0] - yscs[:, 0] * zgse[:, 2],
                            yscs[:, 0] * zgse[:, 1] - yscs[:, 1] * zgse[:, 0]))

    if not isgsetodsl:
        # DSL -> GSE
        dd = data_in[1]
        d0 = dd[:, 0] * my_x[:, 0] + dd[:, 1] * my_y[:, 0] + dd[:, 2] * zgse[:, 0]
        d1 = dd[:, 0] * my_x[:, 1] + dd[:, 1] * my_y[:, 1] + dd[:, 2] * zgse[:, 1]
        d2 = dd[:, 0] * my_x[:, 2] + dd[:, 1] * my_y[:, 2] + dd[:, 2] * zgse[:, 2]
        out_coord = 'GSE'

    else:
        # GSE -> DSL
        dd = data_in[1]
        d0 = dd[:, 0] * my_x[:, 0] + dd[:, 1] * my_x[:, 1] + dd[:, 2] * my_x[:, 2]
        d1 = dd[:, 0] * my_y[:, 0] + dd[:, 1] * my_y[:, 1] + dd[:, 2] * my_y[:, 2]
        d2 = dd[:, 0] * zgse[:, 0] + dd[:, 1] * zgse[:, 1] + dd[:, 2] * zgse[:, 2]
        out_coord = 'DSL'

    dd_out = [d0, d1, d2]
    data_out = np.column_stack(dd_out)

    store_data(name_out, data={'x': data_in[0], 'y': data_out}, attr_dict=meta_copy)
    set_coords(name_out, out_coord)

    return 1



import logging
import warnings
from copy import deepcopy
import numpy as np


def slice2d_collate(data, weight, sphere, previous_out=None, sum_samples=False):
    """
    Collate data aggregated as slice2d_get_data loops over input
    Data aggregation continues until a change in energy or angle
    bins occurs (mode change or other) or aggregation completes.
    At those points this procedure is called to average the data,
    concatenate data to output variables, and clear the appropriate
    variables for the next loop.
    """
    if sum_samples:
        data_out = data
    else:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=RuntimeWarning)
            data_out = data/weight

    data_out = data_out.flatten(order='F')
    weight = weight.flatten(order='F')
    rad_in = sphere['rad'].flatten(order='F')
    phi_in = sphere['phi'].flatten(order='F')
    theta_in = sphere['theta'].flatten(order='F')
    dr_in = sphere['dr'].flatten(order='F')
    dp_in = sphere['dp'].flatten(order='F')
    dt_in = sphere['dt'].flatten(order='F')

    # remove bins with no valid data
    valid = np.argwhere(weight > 0).flatten()

    if len(valid) > 0:
        data_out = data_out[valid]
        rad_in = rad_in[valid]
        phi_in = phi_in[valid]
        theta_in = theta_in[valid]
        dr_in = dr_in[valid]
        dp_in = dp_in[valid]
        dt_in = dt_in[valid]
    else:
        logging.error('No valid data in distribution(s).')
        return

    if previous_out is None:
        rad_out = deepcopy(rad_in)
        phi_out = deepcopy(phi_in)
        theta_out = deepcopy(theta_in)
        dr_out = deepcopy(dr_in)
        dp_out = deepcopy(dp_in)
        dt_out = deepcopy(dt_in)
    else:
        data_out = np.concatenate((previous_out['data'], data_out))
        rad_out = np.concatenate((previous_out['rad'], rad_in))
        phi_out = np.concatenate((previous_out['phi'], phi_in))
        theta_out = np.concatenate((previous_out['theta'], theta_in))
        dr_out = np.concatenate((previous_out['dr'], dr_in))
        dp_out = np.concatenate((previous_out['dp'], dp_in))
        dt_out = np.concatenate((previous_out['dt'], dt_in))

    return {
            'data': data_out,
            'rad': rad_out,
            'phi': phi_out,
            'theta': theta_out,
            'dr': dr_out,
            'dp': dp_out,
            'dt': dt_out
        }



import numpy as np
import unittest
import pyspedas
from pyspedas import tkm2re
from pyspedas.analysis.neutral_sheet import neutral_sheet
from pytplot import get_data

pyspedas.mms.mec()
tkm2re('mms1_mec_r_gsm')
pos_data = get_data('mms1_mec_r_gsm_re')


class NSTests(unittest.TestCase):
    def test_lopez(self):
        model = neutral_sheet(pos_data.times, pos_data.y, model='lopez', sc2NS=True)
        self.assertTrue(isinstance(model, np.ndarray))
        model = neutral_sheet(pos_data.times, pos_data.y, model='lopez')
        self.assertTrue(isinstance(model, np.ndarray))

    def test_sm(self):
        model = neutral_sheet(pos_data.times, pos_data.y, model='sm', sc2NS=True)
        self.assertTrue(isinstance(model, np.ndarray))
        model = neutral_sheet(pos_data.times, pos_data.y, model='sm')
        self.assertTrue(isinstance(model, np.ndarray))

    def test_themis(self):
        model = neutral_sheet(pos_data.times, pos_data.y, model='themis', sc2NS=True)
        self.assertTrue(isinstance(model, np.ndarray))
        model = neutral_sheet(pos_data.times, pos_data.y, model='themis')
        self.assertTrue(isinstance(model, np.ndarray))

    def test_aen(self):
        model = neutral_sheet(pos_data.times, pos_data.y, model='aen', sc2NS=True)
        self.assertTrue(isinstance(model, np.ndarray))
        model = neutral_sheet(pos_data.times, pos_data.y, model='aen')
        self.assertTrue(isinstance(model, np.ndarray))

    def test_den(self):
        model = neutral_sheet(pos_data.times, pos_data.y, model='den', sc2NS=True)
        self.assertTrue(isinstance(model, np.ndarray))
        model = neutral_sheet(pos_data.times, pos_data.y, model='den')
        self.assertTrue(isinstance(model, np.ndarray))

    def test_fairfield(self):
        model = neutral_sheet(pos_data.times, pos_data.y, model='fairfield', sc2NS=True)
        self.assertTrue(isinstance(model, np.ndarray))
        model = neutral_sheet(pos_data.times, pos_data.y, model='fairfield')
        self.assertTrue(isinstance(model, np.ndarray))

    def test_den_fairfield(self):
        model = neutral_sheet(pos_data.times, pos_data.y, model='den_fairfield', sc2NS=True)
        self.assertTrue(isinstance(model, np.ndarray))
        model = neutral_sheet(pos_data.times, pos_data.y, model='den_fairfield')
        self.assertTrue(isinstance(model, np.ndarray))

    def test_invalid_model(self):
        model = neutral_sheet(pos_data.times, pos_data.y, model='ff', sc2NS=True)
        self.assertTrue(not isinstance(model, np.ndarray))


if __name__ == '__main__':
    unittest.main()



from functools import wraps

from .load import load
from .epd.epd import elfin_load_epd

@wraps(elfin_load_epd)
def epd(*args, **kwargs):
    return elfin_load_epd(*args, **kwargs)

def fgm(trange=['2020-10-01', '2020-10-02'],
        probe='a',
        datatype='survey',
        level='l1', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Fluxgate Magnetometer (FGM)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str
            Spacecraft identifier ('a' or 'b')

        datatype: str
            Data type; Valid options:
                'fast', 'survey' for L1 data

        level: str
            Data level; options: 'l1' (default: l1)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='fgm', probe=probe, trange=trange, level=level, 
                datatype=datatype, suffix=suffix, get_support_data=get_support_data, 
                varformat=varformat, varnames=varnames, downloadonly=downloadonly, 
                notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return fgm_postprocessing(tvars)


def fgm_postprocessing(variables):
    """
    Placeholder for FGM post-processing
    """
    return variables





def mrma(trange=['2020-11-5', '2020-11-6'],
        probe='a',
        datatype='mrma', 
        level='l1', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Magneto Resistive Magnetometer (MRMa)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str
            Spacecraft identifier ('a' or 'b')

        datatype: str
            Data type; Valid options:
                'mrma' for L1 data

        level: str
            Data level; options: 'l1' (default: l1)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='mrma', probe=probe, trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return mrma_postprocessing(tvars)


def mrma_postprocessing(variables):
    """
    Placeholder for MRMa post-processing 
    """
    return variables


def mrmi(trange=['2020-11-5', '2020-11-6'],
        probe='a',
        datatype='mrmi', 
        level='l1', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Magneto Resistive Magnetometer (MRMi)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str
            Spacecraft identifier ('a' or 'b')

        datatype: str
            Data type; Valid options:
                'mrmi' for L1 data

        level: str
            Data level; options: 'l1' (default: l1)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='mrmi', probe=probe, trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return mrmi_postprocessing(tvars)


def mrmi_postprocessing(variables):
    """
    Placeholder for MRMi post-processing 
    """
    return variables


def state(trange=['2020-11-5/10:00', '2020-11-5/12:00'],
        probe='a',
        datatype='defn',
        level='l1', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=True):
    """
    This function loads data from the State data (state)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str
            Spacecraft identifier ('a' or 'b')

        datatype: str
            Data type; Valid options:
                'defn' (default)
                'pred'

        level: str
            Data level; options: 'l1' (default: l1)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars= load(instrument='state', probe=probe, trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return state_postprocessing(tvars)


def state_postprocessing(variables):
    """
    Placeholder for State post-processing 
    """
    return variables


def eng(trange=['2020-11-5', '2020-11-6'],
        probe='a',
        datatype='eng_datatype', 
        level='l1', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Engineering (ENG)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str
            Spacecraft identifier ('a' or 'b')

        datatype: str
            Data type; Valid options:
                'eng_datatype' for L1 data

        level: str
            Data level; options: 'l1' (default: l1)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='eng', probe=probe, trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return eng_postprocessing(tvars)


def eng_postprocessing(variables):
    """
    Placeholder for ENG post-processing 
    """
    return variables

        




import logging
import numpy as np
import math
from geopack.geopack import recalc as geopack_recalc
from pyspedas import cotrans, time_double


def sm_ns_model(time, gsm_pos, sc2NS=False):
    """
    This routine calculates the NS position along the zaxis at a specific x and y location.
    """
    # convert gsm to sm coordinates
    sm_pos = cotrans(time_in=time, data_in=gsm_pos, coord_in='gsm', coord_out='sm')
    zns = gsm_pos[:, 2] - sm_pos[:, 2]
    if not sc2NS:
        return zns
    else:
        sc2NS = gsm_pos[:, 2] - zns
        return sc2NS


def themis_ns_model(time, gsm_pos, sc2NS=False):
    """
    NAME:
        themis_ns_model
    PURPOSE:
        This routine calculates the position along the zaxis at a specific
        x and y location. The themis model is used for this calculation.
        The themis model uses z-sm (converted from z-gsm) for the inner probes
        and the Hammond model for the outer probes.
    INPUT:
        time - string or double format
               double(s)  seconds since 1970
               string(s)  format:  YYYY-MM-DD/hh:mm:ss
        gsm_pos - position vector in GSM coordinates in re (pos[*,3])
    OUTPUT: returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS)
            Value is positive if NS is above z=0 gsm plane, negative if below
    KEYWORDS
        sc2NS - if set returns Z displacement from the spacecraft to the neutral sheet
                Value is positive if the NS is northward of the SC location, and negative if below
    NOTES;
        Reference:
        The themis model uses z-sm (converted from z-gsm) for the inner probes
        and the Hammond model (default) for the outer probes. The algorithm can be found
        in ssllib neutralsheet.pro.
    HISTORY:
    """
    # initialize constants and variables
    re = 6378.
    h0 = 8.6   # 10.5 # hinge point of the neutral sheet
    rad = np.pi/180.
    dz2NS = np.zeros(len(time))

    # constants used in hammond model
    H1=8.6
    Y0=20.2
    D=12.2

    # calculate the radial distance
    rdist = np.sqrt(gsm_pos[:,0]**2 + gsm_pos[:,1]**2 + gsm_pos[:,2]**2)

    # Use the sm coordinates for radial distances <= h0  (8.6)
    sm_ind = np.argwhere(rdist <= h0).flatten()
    if len(sm_ind) > 0:
        sm_pos = cotrans(time_in=time[sm_ind], data_in=gsm_pos[sm_ind,:], coord_in='gsm', coord_out='sm')
        dz2NS[sm_ind] = -sm_pos[:, 2]

    # Use the Hammond model for radial distances > h0  (8.6)
    lr_ind = np.argwhere(rdist > h0).flatten()
    if len(lr_ind) > 0:
        # initialize variables
        x = gsm_pos[lr_ind, 0]
        y = gsm_pos[lr_ind, 1]
        z = gsm_pos[lr_ind, 2]
        tilt = np.zeros(len(x))
        # check input time format and convert to doy, hr, min
        for i in range(len(x)):
            # calculate the tilt in degrees
            tilt[i] = geopack_recalc(time_double(time[lr_ind[i]]))

        # hammond model
        iless = np.argwhere(np.abs(y) < Y0).flatten()
        if len(iless) > 0:
            dz2NS[lr_ind[iless]] = ((H1+D)*np.sqrt(1-y[iless]**2/Y0**2)-D)*np.sin(tilt[iless])

        imore = np.argwhere(np.abs(y) >= Y0).flatten()
        if len(imore) > 0:
            dz2NS[lr_ind[imore]] = -D*np.sin(tilt[imore])

    if not sc2NS:
        return gsm_pos[:, 2] - (-dz2NS)
    else:
        return -dz2NS


def aen_ns_model(time, gsm_pos, sc2NS=False):
    """
    NAME:
    aen_ns_model

    PURPOSE:  This program is to find the AEN(Analytical Equatorial Neutral) sheet in the
              magnetopause in different time and position

    INPUT:
    time - string or double format
           double(s)  seconds since 1970
           string(s)  format:  YYYY-MM-DD/hh:mm:ss
    gsm_pos - position vector in GSM coordinates in re (pos[*,3])

    OUTPUT: returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS)
            Value is positive if NS is above z=0 gsm plane, negative if below

    KEYWORDS
        sc2NS - if set returns Z displacement from the spacecraft to the neutral sheet
                Value is positive if the NS is northward of the SC location, and negative if below

    NOTES:

    References:
    (1) AEN(Analytical Equatorial Neutral):
        Zhu, M. and R.-L. Xu, 1994, A continuous neutral sheet model and a normal
        curved coordinate system in the magnetotail,  Chinese J. Space Science,  14,
        (4)269, (in Chinese).
        Wang, Z.-D. and R.-L. Xu, Neutral Sheet Observed on ISEE Satellite,
        Geophysical Research Letter, 21, (19)2087, 1994.
    (2) Magnetopause model:
        Sibeck, D. G., R. E. Lopez, and E. C. Roelof, Solar wind control of the
        magnetopause shape, location, and motion, J. Grophys. Res., 96, 5489, 1991

    HISTORY:

    """
    # initialize constants
    h0 = 12.6/np.pi

    dz2ns = np.zeros(len(time))

    for i in range(len(time)):
        # calculate the tilt angle
        tt = geopack_recalc(time_double(time[i]))

        # calculate the position of the neutral sheet
        dz2ns[i] = -h0 * np.sin(tt) * np.arctan(gsm_pos[i,0]/5) * (2*np.cos(gsm_pos[i,1]/6))

    if not sc2NS:
        return dz2ns
    else:
        return gsm_pos[:, 2]-dz2ns


def den_ns_model(time, gsm_pos, sc2NS=False):
    """
    NAME:
        den_ns_model
    PURPOSE:
        This program finds the DEN(Displaced Equatorial Neutral) sheet inside
        the magnetopause in different tine and positions. The routine calculates
        the position along the zaxis at a specific location.
    INPUT:
        time - string or double format
               double(s)  seconds since 1970
               string(s)  format:  YYYY-MM-DD/hh:mm:ss
        gsm_pos - position vector in GSM coordinates in re (pos[*,3])
    OUTPUT: returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS)
            Value is positive if NS is above z=0 gsm plane, negative if below
    KEYWORDS
        sc2NS - if set returns Z displacement from the spacecraft to the neutral sheet
                Value is positive if the NS is northward of the SC location, and negative if below
    NOTES:
        References:
        (1) DEN(Displaced Equatorial Neutral):
            Xu, R.-L., A Displaced Equatorial Neutral Sheet Surface Observed on ISEE-2
            Satellite,  J. Atmospheric and Terrestrial Phys., 58, 1085, 1991
        (2) Magnetopause model:
            Sibeck, D. G., R. E. Lopez, and R. C. Roelof, Solar wind control of the
            magnetopause shape, location, and motion, J. Grophys. Res., 96, 5489, 1991
        Original Authors of the FORTRAN source code:
        Ronglan XU and Lei LI, Center for Space Sci. and Applied Res.,
        Chinese Academy of Sciences, PO Box 8701, Beijing 100080, China
        E-mail: XURL@SUN.IHEP.AC.CN, XURL@SUN20.CSSAR.AC.CN
        This source code was ported from the original FORTRAN source code into IDL
        The original source code only calculated to 10.05 RE. In this IDL version
        that restriction was increased to 25.
    HISTORY:
    """
    # calculate the position of the neutral sheet along z axis
    H = 25.5
    H1 = 25.05

    dz2ns = np.zeros(len(time))

    for i in range(len(time)):
        done = 0
        xgsm = gsm_pos[i,0]
        ygsm = gsm_pos[i,1]

        # get tilt angle of magnetic pole
        tilt = geopack_recalc(time_double(time[i]))

        if xgsm > -100.:
            d = sd1(tilt, H, H1, xgsm)
            ym21 = ((H1*(H+d))**2) * (1-(xgsm/(H*np.cos(tilt)))**2)
            ym22 = (H+d)**2 - (d-xgsm/np.cos(tilt))**2
            ym2 = ym21/ym22
            if ym2 < 0:
                #ie[i] = 2
                continue
            ym = np.sqrt(ym2)
            xd2 = ((H*np.cos(tilt))**2) * (1-(ygsm/H1)**2)
            if np.abs(ygsm) > H1:
                xd2 = 0
            # find the equatorial region
            xd = np.sqrt(xd2)
            rd = np.sqrt(xd**2+ygsm**2)
            rsm = np.sqrt(xgsm**2+ygsm**2)
            if xgsm > 0 or rsm <= rd:
                dz2ns[i] = -xgsm*np.sin(tilt)/np.cos(tilt)
                done = 1
            if np.abs(ygsm) > ym and done != 1:
                dz2ns[i] = -d*np.sin(tilt)
                done = 1
            if done != 1:
                dz2ns[i] = ((H+d)*np.sqrt(1-(ygsm**2)/ym2)-d)*np.sin(tilt)

    if not sc2NS:
        return dz2ns
    else:
        sc2NS = gsm_pos[:,2] - dz2ns
        return sc2NS


def sfa4(aa, bb, cc, dd):
    ndx = 0
    xmin = 0
    xmax = 50
    ndxmax = 3
    dx = 1
    x = xmin
    yy = x**4 + aa*x**3 + bb*x**2 + cc*x + dd

    while ndx <= ndxmax:
        x = x+dx
        if x >= xmax:
            ndx = 0
            return x
        y = x**4 + aa*x**3 + bb*x**2 + cc*x + dd
        ry = y/yy
        if ry < 0:
            x = x-dx
            dx = dx/10.
            ndx = ndx+1
        else:
            yy = y
    return x


def sd1(til, H, H1, xgsm):
    ct = np.cos(til)
    xx = xgsm
    xh = -H*ct
    if xgsm >= xh:
        xx = xh

    # calculate the radius of the cross section
    if xx <= -5.:
        rm = 9*(10-3*xx)/(10-xx)+3
    if xx > -5.:
        rm = np.sqrt(18**2-(xx+5)**2)
    rm2 = rm**2

    # in cross_section areas above and below the neutral
    # sheet
    aa = 4*H-(32*rm2*H**2)/(np.pi**2*H1**2*(H-xx/ct))
    bb = 2*H**2*(3.-8.*rm2/(np.pi**2*H1**2))
    cc = 4*(H**3)
    dd = H**4
    x = sfa4(aa, bb, cc, dd)

    d = x
    if xgsm >= xh:
        fk = -x/np.sqrt(-xh)
        d = -fk*np.sqrt(-xgsm)
    return d


def fairfield_ns_model(time, gsm_pos, sc2NS=False):
    """
    NAME:
    fairfield_NS_model

    PURPOSE:
    This routine calculates the position along the zaxis at a specific
    x and y location. The Fairfield model is used to this calculation.

    INPUT:
    time - string or double format
           double(s)  seconds since 1970
           string(s)  format:  YYYY-MM-DD/hh:mm:ss
    gsm_pos - position vector in GSM coordinates in re (pos[*,3])

    OUTPUT: returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS)
            Value is positive if NS is above z=0 gsm plane, negative if below

    KEYWORDS
        sc2NS - if set returns Z displacement from the spacecraft to the neutral sheet
                Value is positive if the NS is northward of the SC location, and negative if below

    NOTES:
    Reference:
    A statistical determination of the shape and position of the
    geomagnetic neutral sheet,  Journal of Geophysical Research,
    Vol. 85, No A2, pages 775-780, February 1, 1980
    Author - D. Fairfield

    HISTORY:
    """

    # constants (in re)
    h0 = 10.5
    y0 = 22.5
    d = 14.

    dz2NS = np.zeros(len(time))
    tilt = np.zeros(len(time))

    for i in range(len(time)):
        # calculate tilt angle of geomagnetic axis
        tilt[i] = geopack_recalc(time_double(time[i]))

    # calculate the position of the neutral sheet along z axis
    y_ge_y0 = np.argwhere(np.abs(gsm_pos[:,0]) >= y0).flatten()
    y_lt_y0 = np.argwhere(np.abs(gsm_pos[:,0]) < y0).flatten()
    if len(y_ge_y0) > 0:
        dz2NS[y_ge_y0] = -d*np.sin(tilt[y_ge_y0])
    if len(y_lt_y0) > 0:
        dz2NS[y_lt_y0] = ((h0 + d) * np.sqrt(1 - gsm_pos[y_lt_y0,0]**2/y0**2) - d)*np.sin(tilt[y_lt_y0])

    if not sc2NS:
        return dz2NS
    else:
        sc2NS = gsm_pos[:, 2] - dz2NS
        return sc2NS

   
def den_fairfield_ns_model(time, gsm_pos, sc2NS=False):
    """
    NAME:
     den_fairfield_ns_model
    PURPOSE:
     This routine calculates the position along the zaxis at a specific
     x and y location.
    INPUT:
     time - string or double format
            double(s)  seconds since 1970
            string(s)  format:  YYYY-MM-DD/hh:mm:ss
     gsm_pos - position vector in GSM coordinates in re (pos[*,3])
    OUTPUT: returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS)
            Value is positive if NS is above z=0 gsm plane, negative if below
    KEYWORDS
     sc2NS - if set returns Z displacement from the spacecraft to the neutral sheet
             Value is positive if the NS is northward of the SC location, and negative if below
    HISTORY:
    """
    # initialize constants
    dz2ns = np.zeros(len(time))

    # Use the den model for radial distances <12.re
    rdist = np.sqrt(gsm_pos[:,0]**2 + gsm_pos[:,1]**2 + gsm_pos[:,2]**2)
    sm_ind = np.argwhere(rdist <= 10.).flatten()
    if len(sm_ind) > 0:
        dz2ns[sm_ind] = den_ns_model(time[sm_ind], gsm_pos[sm_ind,:])

    # use the fairfield model for radial distances >12.re
    lr_ind = np.argwhere(rdist > 10.).flatten()
    if len(lr_ind) > 0:
        dz2ns[lr_ind] = fairfield_ns_model(time[lr_ind], gsm_pos[lr_ind,:])

    if not sc2NS:
        return dz2ns
    else:
        sc2NS = gsm_pos[:,2] - dz2ns
        return sc2NS


def lopez_ns_model(time, gsm_pos, kp=None, mlt=None, sc2NS=False):
    """
    NAME:
        lopez_ns_model
    PURPOSE:
        This routine calculates the position along the zaxis at a specific
        x and y location. The Lopez model is used for this calculation.
    INPUT:
        time - string or double format
               double(s)  seconds since 1970
               string(s)  format:  YYYY-MM-DD/hh:mm:ss
        gsm_pos - position vector in GSM coordinates in re (pos[*,3])
        kp - kp index value
        mlt - magnetic local time in degrees (0=midnight)
    OUTPUT: returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS)
            Value is positive if NS is above z=0 gsm plane, negative if below
    KEYWORDS
        sc2NS - if set returns Z displacement from the spacecraft to the neutral sheet
                Value is positive if the NS is northward of the SC location, and negative if below
    NOTES:
        Reference:
        The position of the magnetotail neutral sheet in the near-Earth Region,
        Geophysical Research Letters, Vol. 17, No 10, pages 1617-1620, 1990
        Author - Ramon E. Lopez
        The lopez model is best used for distances <8.8 RE
    HISTORY:
    """
    # constants
    rad = np.pi/180.
    if kp is None:
        kp = 0
    if mlt is None:
        mlt = 0.0
    tilt = np.zeros(len(time))

    for i in range(len(time)):
        # calculate tilt angle of geomagnetic axis
        tilt[i] = geopack_recalc(time[i])

    # calculate the position of the neutral sheet along z axis
    rdist = np.sqrt(gsm_pos[:,0]**2 + gsm_pos[:,1]**2 + gsm_pos[:,2]**2)
    mlat = -(0.14*kp + 0.69) * ((np.cos(rad*mlt))**.3333333) * (0.065*(rdist**0.8) - 0.16) * tilt * 180.0/np.pi
    mlat = mlat + tilt * 180.0/np.pi

    # convert magnetic latitude to position
    x, y, z = rthph2xyz(rdist, mlat, mlt)

    if not sc2NS:
        return z
    else:
        sc2NS = gsm_pos[:,2] - z
        return sc2NS


def rthph2xyz(r,th,ph):
    """
    Helper function for the lopez model
    converts spherical to cartesian coordinates
    NOTE: th,ph in degrees, and th is latitude (not colatitude) (i.e. [-90->90])
    """
    FLAG=6.8792E+28
    FLAG98=0.98*FLAG
    PI=3.1415926535898

    thrad=th*PI/180.
    phrad=ph*PI/180.
    sth=np.sin(thrad)
    cth=np.cos(thrad)
    sph=np.sin(phrad)
    cph=np.cos(phrad)
    x=r*cth*cph
    y=r*cth*sph
    z=r*sth

    iflags=np.argwhere((r > FLAG98) | (th > FLAG98) | (ph > FLAG98)).flatten()
    if (len(iflags) > 0):
        x[iflags]=FLAG
        y[iflags]=FLAG
        z[iflags]=FLAG

    return x,y,z


"""
C========================================================================================
C
      SUBROUTINE TAG14_EQUAT_SHEET (XGSM,YGSM,ZGSM,PSI,PDYN,ByIMF,BzIMF,
     * XGSM_S,YGSM_S,ZGSM_S)
C
C  INPUT:   XGSM,YGSM,ZGSM: position of a point inside the magnetosphere
C           PSI:  geodipole tilt angle in radians
C           PDYN: solar wind ram pressure in nanoPascals
c           ByIMF, BzIMF:  Y- and Z- GSM components of the IMF in nT
c                            (averaged over previous 30 minutes)
C
c  OUTPUT:  XGSM_S,YGSM_S,ZGSM_S: GSM coordinates of a point of the TAG14 equatorial sheet,
c           located at the same geocentric distance R=sqrt(XGSM^2+YGSM^2+ZGSM^2)
c           and lying in the same GSM meridian plane as the original point {XGSM,YGSM,ZGSM}
c
c  Author: N. A. Tsyganenko, Dec.18, 2014
c  Reference:  Tsyganenko, N. A., V. A. Andreeva, and E. I. Gordeev, Internally and externally induced deformations
c              of the magnetospheric equatorial current as inferred from spacecraft data, 
c              Ann.Geophys., v.33, pp.1-11, doi:10.5194/angeo-33-1-2015, 2015.
c

Translated from FORTRAN via Chat-GPT4. 
Original FORTRAN source: https://geo.phys.spbu.ru/~tsyganenko/models/cs/TAG14.for

"""


def tag14_equat_sheet(xgsm, ygsm, zgsm, psi, pdyn, byimf, bzimf):
    PI = 3.1415926539
    ERR = 1e-8

    # Constants
    RH0, RH1, RH2, RH3, RH4, RH5, T0, T1, A00, A01, A02, A10, A11, A12, ALPHA0, DALPHA1, DALPHA2, DALPHA3, XAPPA, BETA0, BETA1 = (
        11.01595255, 6.054799445, 0.8376136340, -2.283371357, -0.2505011095, -0.9648465280, 0.2865647076, 0.1847283161,
        2.909423481, -0.1585372900, 0.5648121327, 1.889466989, 0.06086006001, 0.4853873874, 7.128509839, 4.867577218,
        -0.2248219212, -0.1441988549, -0.2868279445, 2.179888855, 0.4035940369
    )

    xgsm1 = xgsm
    ygsm1 = ygsm
    zgsm1 = zgsm
    r = math.sqrt(xgsm1 ** 2 + ygsm1 ** 2 + zgsm1 ** 2)
    theta = math.acos(zgsm1 / r)

    if xgsm1 == 0 and ygsm1 == 0:
        phi = 0
    else:
        phi = math.atan2(ygsm1, xgsm1)

    byfact = byimf / 5.0
    bzfact = bzimf / 5.0

    pdyn_0 = 2.0
    pfact = (pdyn / pdyn_0) ** XAPPA - 1.0

    cps = math.cos(psi)
    sps = math.sin(psi)
    cp = math.cos(phi)
    sp = math.sin(phi)
    th1 = 0.0
    th2 = PI

    while abs(th1 - th2) > ERR:
        th = 0.5 * (th1 + th2)

        xgsm1 = r * math.sin(th) * cp
        ygsm1 = r * math.sin(th) * sp
        zgsm1 = r * math.cos(th)
        xsm = xgsm1 * cps - zgsm1 * sps
        ysm = ygsm1
        zsm = xgsm1 * sps + zgsm1 * cps

        rho = math.sqrt(xsm ** 2 + ysm ** 2)

        if abs(rho) < 1e-6:
            cosphi = 1.0
            sinphi = 0.0
        else:
            cosphi = xsm / rho
            sinphi = ysm / rho

        alpha = ALPHA0 + DALPHA1 * cosphi + DALPHA2 * pfact + DALPHA3 * bzfact
        beta = BETA0 + BETA1 * bzfact

        rh = RH0 + RH1 * pfact + RH2 * bzfact + (RH3 + RH4 * pfact + RH5 * bzfact) * cosphi
        t = T0 + T1 * pfact
        a0 = A00 + A01 * pfact + A02 * bzfact
        a1 = A10 + A11 * pfact + A12 * bzfact

        f = 1.0 - (1.0 + (rho / rh) ** alpha) ** (1.0 / alpha)
        g = a0 + a1 * cosphi
        f1 = t * byfact * (rho / 10.0) ** beta * sinphi

        zsm_sheet = rh * math.tan(psi) * f * g + f1

        ff = zsm - zsm_sheet

        if ff > 0:
            th1 = th
        else:
            th2 = th

    theta_s_gsm = 0.5 * (th1 + th2)

    xgsm_s = r * math.sin(theta_s_gsm) * cp
    ygsm_s = r * math.sin(theta_s_gsm) * sp
    zgsm_s = r * math.cos(theta_s_gsm)

    return xgsm_s, ygsm_s, zgsm_s



def neutral_sheet(time, pos, kp=None, model='themis', mlt=None, in_coord='gsm', sc2NS=False):
    """
    Calculate the distance to the neutral sheet for a given time and position.

    Parameters
    ----------
    time : datetime
        Time of interest.
    pos : array_like
        Position of interest.
    kp : array_like, optional
        Kp index.
    model : str, optional
        Neutral sheet model to use.
    mlt : array_like, optional
        Magnetic local time.
    in_coord : str, optional
        Coordinate system of the input position.
    sc2NS : Bool, optional
        Flag to return spacecraft to neutral sheet distance.

    Returns
    -------
    distance2NS : array_like
        Distance to the neutral sheet.

    """

    time = np.array(time)

    # validate and initialize parameters if not set
    if model is None:
        model = 'themis'
    else:
        model = model.lower()
    models = ['sm', 'themis', 'aen', 'den', 'fairfield', 'den_fairfield', 'lopez']
    if model not in models:
        logging.error('An invalid neutral sheet model name was used. Valid entries include: ')
        logging.error(models)
        return

    # check input coordinate system, convert to gsm if needed
    if in_coord is None:
        in_coord = 'gsm'
    else:
        in_coord = in_coord.lower()
    if in_coord == 'gsm':
        gsm_pos = pos
    else:
        gsm_pos = cotrans(time_in=time, data_in=pos, coord_in=in_coord, coord_out='gsm')

    # call the appropriate neutral sheet model
    if model == 'sm':
        return sm_ns_model(time, gsm_pos, sc2NS=sc2NS)
    elif model == 'themis':
        return themis_ns_model(time, gsm_pos, sc2NS=sc2NS)
    elif model == 'aen':
        return aen_ns_model(time, gsm_pos, sc2NS=sc2NS)
    elif model == 'den':
        return den_ns_model(time, gsm_pos, sc2NS=sc2NS)
    elif model == 'fairfield':
        return fairfield_ns_model(time, gsm_pos, sc2NS=sc2NS)
    elif model == 'den_fairfield':
        return den_fairfield_ns_model(time, gsm_pos, sc2NS=sc2NS)
    elif model == 'lopez':
        return lopez_ns_model(time, gsm_pos, kp=kp, mlt=mlt, sc2NS=sc2NS)






from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pytplot import time_clip as tclip
from pytplot import cdf_to_tplot

from .config import CONFIG

def load(trange=['1997-01-03', '1997-01-04'], 
         instrument='mfe',
         datatype='k0',
         suffix='', 
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the Polar mission; this function is not meant 
    to be called directly; instead, see the wrappera:
        pyspedas.polar.mfe
        pyspedas.polar.efi
        pyspedas.polar.pwi
        pyspedas.polar.hydra
        pyspedas.polar.tide
        pyspedas.polar.timas
        pyspedas.polar.cammice
        pyspedas.polar.ceppad
        pyspedas.polar.uvi
        pyspedas.polar.vis
        pyspedas.polar.pixie
        pyspedas.polar.orbit

    """

    if instrument == 'mfe':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'efi':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'pwi':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'hydra':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_hyd_%Y%m%d_v??.cdf'
    elif instrument == 'tide':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_tid_%Y%m%d_v??.cdf'
    elif instrument == 'timas':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_tim_%Y%m%d_v??.cdf'
    elif instrument == 'cammice':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_cam_%Y%m%d_v??.cdf'
    elif instrument == 'ceppad':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_cep_%Y%m%d_v??.cdf'
    elif instrument == 'uvi':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'vis':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'pixie':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_pix_%Y%m%d_v??.cdf'
    elif instrument == 'spha':
        pathformat = 'orbit/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars




import logging
import pytplot

def tcrossp(v1, v2, newname=None, return_data=False):
    """
    Calculates the cross product of two tplot varibles

    Input
    -------
        v1: str
            First tplot variable

        v2: str
            Second tplot variable

    Parameters
    -----------
        newname: str
            Name of the output variable

        return_data: bool
            Returns the data as an ndarray instead of creating a tplot variable

    Returns
    --------
        Name of the tplot variable
    """
    logging.info("tcrossp has been moved to the pytplot.tplot_math module. Please update your imports!")
    logging.info("This version will eventually be removed.")
    return pytplot.tplot_math.tcrossp(v1=v1, v2=v2, newname=newname,return_data=return_data)


import cdflib
import numpy as np

from pytplot import get_data, store_data, options, clip, ylim

from ...satellite.erg.load import load


def gmag_mm210(
    trange=['2020-08-01', '2020-08-02'],
    suffix='',
    site='all',
    datatype='all',
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    uname=None,
    passwd=None,
    time_clip=False,
    ror=True
):

    site_code_all = ['tik', 'zgn', 'yak', 'irt', 'ppi', 'bji',
                    'lnp', 'mut', 'ptn', 'wtk', 'lmt', 'kat',
                    'ktn', 'chd', 'zyk', 'mgd', 'ptk', 'msr',
                    'rik', 'onw', 'kag', 'ymk', 'cbi', 'gua',
                    'yap', 'kor', 'ktb', 'bik', 'wew', 'daw',
                    'wep', 'bsv', 'dal', 'can', 'adl', 'kot',
                    'cst', 'ewa', 'asa', 'mcq']
    tres_all=['1sec', '1min', '1h']
    if isinstance(datatype, str):
        datatype = datatype.lower()
        datatype = datatype.split(' ')
    elif isinstance(datatype, list):
        for i in range(len(datatype)):
            datatype[i] = datatype[i].lower()

    if 'all' in datatype:
        datatype=tres_all
    datatype = list(set(datatype).intersection(tres_all))
    if len(datatype) < 1:
        return

    if '1s' in datatype:
        index = np.where(np.array(datatype) == '1s')[0][0]
        datatype[index] = '1sec'
    elif  '1m' in datatype:
        index = np.where(np.array(datatype) == '1m')[0][0]
        datatype[index] = '1min'
    elif  '1hr' in datatype:
        index = np.where(np.array(datatype) == '1hr')[0][0]
        datatype[index] = '1h'

    
    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(' ')
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if 'all' in site_code:
        site_code = site_code_all
    
    site_code = list(set(site_code).intersection(site_code_all))

    prefix = 'mm210_'
    if notplot:
        loaded_data = {}
    else:
        loaded_data = []
    for site_input in site_code:
        for data_type_in in datatype:
            fres = data_type_in

            if fres == '1h':
                fres = '1min'

            file_res = 3600. * 24
            pathformat = 'ground/geomag/mm210/'+fres+'/'+site_input\
                            +'/%Y/mm210_'+fres+'_'+site_input+'_%Y%m%d_v??.cdf'
            
            loaded_data_temp = load(pathformat=pathformat, file_res=file_res, trange=trange, datatype=datatype, prefix=prefix, suffix='_'+site_input+suffix, get_support_data=get_support_data,
                            varformat=varformat, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd)
            
            if notplot:
                loaded_data.update(loaded_data_temp)
            else:
                loaded_data += loaded_data_temp
            if (len(loaded_data_temp) > 0) and ror:
                try:
                    if isinstance(loaded_data_temp, list):
                        if downloadonly:
                            cdf_file = cdflib.CDF(loaded_data_temp[-1])
                            gatt = cdf_file.globalattsget()
                        else:
                            gatt = get_data(loaded_data_temp[-1], metadata=True)['CDF']['GATT']
                    elif isinstance(loaded_data_temp, dict):
                        gatt = loaded_data_temp[list(loaded_data_temp.keys())[-1]]['CDF']['GATT']
                    print('**************************************************************************')
                    print(gatt["Logical_source_description"])
                    print('')
                    print(f'Information about {gatt["Station_code"]}')
                    print('PI and Host PI(s):')
                    print(gatt["PI_name"])
                    print('')
                    print('Affiliations: ')
                    print(gatt["PI_affiliation"])
                    print('')
                    print('Rules of the Road for 210 MM Data Use:')
                    for gatt_text in gatt["TEXT"]:
                        print(gatt_text)
                    print(f'{gatt["LINK_TEXT"]} {gatt["HTTP_LINK"]}')
                    print('**************************************************************************')
                except:
                    print('printing PI info and rules of the road was failed')
                
            if (not downloadonly) and (not notplot):
                if fres == '1min':
                    fres_list = ['1min', '1h']
                else:
                    fres_list = [fres]
                for fres_in in fres_list:
                    current_tplot_name = prefix+'hdz_'+fres_in+'_' + site_input+suffix
                    if current_tplot_name in loaded_data:
                        get_data_vars = get_data(current_tplot_name)
                        if get_data_vars is None:
                            store_data(current_tplot_name, delete=True)
                        else:
                            #;--- Rename
                            new_tplot_name = prefix+'mag_'+site_input+'_'+fres_in+'_hdz'+suffix
                            store_data(current_tplot_name, newname=new_tplot_name)
                            loaded_data.remove(current_tplot_name)
                            loaded_data.append(new_tplot_name)
                            #;--- Missing data -1.e+31 --> NaN
                            clip(new_tplot_name, -1e+4, 1e+4)
                            get_data_vars = get_data(new_tplot_name)
                            ylim(new_tplot_name, np.nanmin(get_data_vars[1]), np.nanmax(get_data_vars[1]))
                            #;--- Labels
                            options(new_tplot_name, 'legend_names', ['Ch1','Ch2','Ch3'])
                            options(new_tplot_name, 'Color', ['b', 'g', 'r'])
                            options(new_tplot_name, 'ytitle', '\n'.join(new_tplot_name.split('_')))


    return loaded_data



from .load import load

def sspc(trange=['2013-01-28', '2013-01-29'],
        probe='1A',
        downloadonly=False,
        no_update=False,
        time_clip=False):
   
    return load(datatype='sspc', trange=trange, probe=probe, downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)

def mspc(trange=['2013-01-29','2013-01-30'],
        probe='1A',
        downloadonly=False,
        no_update=False,
        time_clip=False):
   
    return load(datatype='mspc', trange=trange, probe=probe, downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)

def fspc(trange=['2013-01-29','2013-01-30'],
        probe='1A',
        downloadonly=False,
        no_update=False,
        time_clip=False):
   
    return load(datatype='fspc',trange=trange, probe=probe, downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)

def rcnt(trange=['2013-01-29','2013-01-30'],
        probe='1A',
        downloadonly=False,
        no_update=False,
        time_clip=False):
   
    return load(datatype='rcnt', trange=trange, probe=probe, downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)

def magn(trange=['2013-01-29','2013-01-30'],
        probe='1A',
        downloadonly=False,
        no_update=False,
        time_clip=False):
   
    return load(datatype='magn', trange=trange, probe=probe, downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)

def ephm(trange=['2013-01-29','2013-01-30'],
        probe='1A',
        downloadonly=False,
        no_update=False,
        time_clip=False):
   
    return load(datatype='ephm', trange=trange, probe=probe, downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)

def hkpg(trange=['2013-01-29','2013-01-30'],
        probe='1A',
        downloadonly=False,
        no_update=False,
        time_clip=False):
   
    return load(datatype='hkpg', trange=trange, probe=probe, downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)



import os
import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):

    def test_load_fgm_data(self):
        out_vars = pyspedas.elfin.fgm(time_clip=True)
        self.assertTrue(data_exists('ela_fgs'))

    def test_load_epd_data(self):
        out_vars = pyspedas.elfin.epd()
        self.assertTrue(data_exists('ela_pef'))

    def test_load_mrma_data(self):
        out_vars = pyspedas.elfin.mrma()
        self.assertTrue(data_exists('ela_mrma'))

    def test_load_mrmi_data(self):
        out_vars = pyspedas.elfin.mrmi()
        self.assertTrue(data_exists('ela_mrmi'))

    def test_load_state_data(self):
        out_vars = pyspedas.elfin.state()
        self.assertTrue(data_exists('ela_pos_gei'))
        self.assertTrue(data_exists('ela_vel_gei'))

    def test_load_eng_data(self):
        out_vars = pyspedas.elfin.eng()
        self.assertTrue(data_exists('ela_fc_idpu_temp'))

    def test_load_notplot(self):
        out_vars = pyspedas.elfin.epd(notplot=True)
        self.assertTrue('ela_pef' in out_vars)

    def test_downloadonly(self):
        files = pyspedas.elfin.epd(downloadonly=True, trange=['2020-11-01', '2020-11-02'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()

    


import numpy as np


def minvar(data):
    """
    This program computes the principal variance directions and variances of a
    vector quantity as well as the associated eigenvalues.

    Parameters
    -----------
    data:
        Vxyz, an (npoints, ndim) array of data(ie Nx3)

    Returns
    -------
    vrot:
        an array of (npoints, ndim) containing the rotated data in the new coordinate system, ijk.
        Vi(maximum direction)=vrot[0,:]
        Vj(intermediate direction)=vrot[1,:]
        Vk(minimum variance direction)=Vrot[2,:]
    v:
        an (ndim,ndim) array containing the principal axes vectors
        Maximum variance direction eigenvector, Vi=v[*,0]
        Intermediate variance direction, Vj=v[*,1] (descending order)
    w:
        the eigenvalues of the computation
    """

    #  Min var starts here
    # data must be Nx3
    vecavg = np.nanmean(np.nan_to_num(data, nan=0.0), axis=0)

    mvamat = np.zeros((3, 3))
    for i in range(3):
        for j in range(3):
            mvamat[i, j] = np.nanmean(np.nan_to_num(data[:, i] * data[:, j], nan=0.0)) - vecavg[i] * vecavg[j]

    # Calculate eigenvalues and eigenvectors
    w, v = np.linalg.eigh(mvamat, UPLO='U')

    # Sorting to ensure descending order
    w = np.abs(w)
    idx = np.flip(np.argsort(w))

    # IDL compatability
    if True:
        if np.sum(w) == 0.0:
            idx = [0, 2, 1]

    w = w[idx]
    v = v[:, idx]

    # Rotate intermediate var direction if system is not Right Handed
    YcrossZdotX = v[0, 0] * (v[1, 1] * v[2, 2] - v[2, 1] * v[1, 2])
    if YcrossZdotX < 0:
        v[:, 1] = -v[:, 1]
        # v[:, 2] = -v[:, 2] # Should not it is being flipped at Z-axis?

    # Ensure minvar direction is along +Z (for FAC system)
    if v[2, 2] < 0:
        v[:, 2] = -v[:, 2]
        v[:, 1] = -v[:, 1]

    vrot = np.array([np.dot(row, v) for row in data])

    return vrot, v, w



"""
Creates a new tplot variable that has spikes removed.

Notes
-----
Similar to clean_spikes.pro in IDL SPEDAS.

"""
import logging
import pytplot


def clean_spikes(names, nsmooth=10, thresh=0.3, sub_avg=False,
                 new_names=None, suffix=None, overwrite=None):
    """
    Clean spikes from data.

    Parameters
    ----------
    names: str/list of str
        List of pytplot names.
    new_names: str/list of str, optional
        List of new_names for pytplot variables.
        If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply. Default is '-avg'.
    overwrite: bool, optional
        Replace the existing tplot name.
    nsmooth: int, optional
        the number of data points for smoothing
    thresh: float, optional
        threshold value
    sub_avg: bool, optional
        if set, subtract the average value of the data
        prior to checking for spikes

    Returns
    -------
    None.

    """
    logging.info("clean_spikes has been moved to the pytplot.tplot_math package. Please update your imports!")
    logging.info("This version will eventually be removed.")
    pytplot.tplot_math.clean_spikes(names=names, nsmooth=nsmooth, thresh=thresh, sub_avg=sub_avg, new_names=new_names,
                                    suffix=suffix,overwrite=overwrite)



from pyspedas.elfin.load import load

def state(trange=['2020-11-5/10:00', '2020-11-5/12:00'],
          probe='a',
          datatype='defn',
          level='l1',
          suffix='',
          get_support_data=False,
          varformat=None,
          varnames=[],
          downloadonly=False,
          notplot=False,
          no_update=False,
          time_clip=True):
    """
    This function loads data from the State data (state)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str
            Spacecraft identifier ('a' or 'b')

        datatype: str
            Data type; Valid options:
                'defn' (default)
                'pred'

        level: str
            Data level; options: 'l1' (default: l1)

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='state', probe=probe, trange=trange, level=level, datatype=datatype, suffix=suffix,
                 get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly,
                 notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return state_postprocessing(tvars)


def state_postprocessing(variables):
    """
    Placeholder for State post-processing
    """
    return variables




import numpy as np

# use nansum from bottleneck if it's installed, otherwise use the numpy one
try:
    import bottleneck as bn
    nansum = bn.nansum
except ImportError:
    nansum = np.nansum


def ctv_err_test(vals, val, err = 1e-5):
    """ used to determine if values are equal within some standard of computational error """
    return (vals >= val-err) & (vals <= val+err)


def ctv_determ_mats(m):
    """ returns the determinant of a list of 3x3 matrices """
    return np.linalg.det(m)


def ctv_identity_mats(m):
    """
    determines if a list of 3x3 matrices are identity matrices
    will return the indexes of the identity matrices in the list of matrices
    """
    return ctv_err_test(m[:, 0, 0], 1) & ctv_err_test(m[:, 0, 1], 0) & ctv_err_test(m[:, 0, 2], 0) \
        & ctv_err_test(m[:, 1, 0], 0) & ctv_err_test(m[:, 1, 1], 1) & ctv_err_test(m[:, 1, 2], 0) \
        & ctv_err_test(m[:, 2, 0], 0) & ctv_err_test(m[:, 2, 1], 0) & ctv_err_test(m[:, 2, 2], 1)


def ctv_mm_mult(m1, m2):
    """ multiplication of two lists of 3x3 matrices """
    out = np.zeros(m1.shape)
    out[:, 0, 0] = nansum(m1[:, 0, :] * m2[:, :, 0], axis=1)
    out[:, 1, 0] = nansum(m1[:, 1, :] * m2[:, :, 0], axis=1)
    out[:, 2, 0] = nansum(m1[:, 2, :] * m2[:, :, 0], axis=1)
    out[:, 0, 1] = nansum(m1[:, 0, :] * m2[:, :, 1], axis=1)
    out[:, 1, 1] = nansum(m1[:, 1, :] * m2[:, :, 1], axis=1)
    out[:, 2, 1] = nansum(m1[:, 2, :] * m2[:, :, 1], axis=1)
    out[:, 0, 2] = nansum(m1[:, 0, :] * m2[:, :, 2], axis=1)
    out[:, 1, 2] = nansum(m1[:, 1, :] * m2[:, :, 2], axis=1)
    out[:, 2, 2] = nansum(m1[:, 2, :] * m2[:, :, 2], axis=1)
    return out


def ctv_verify_mats(m):
    """
    verifies whether a list of matrices
    contains valid rotation matrices.
    This is determined using 2 constraints.
    #1 Where determ(matrix) eq 1
    #2 Where matrix#transpose(matrix) eq I

    returns 0 if the matrices use a mixed system
    returns 1 if there are no valid mats
    returns 2 if the data are all nans
    returns 3 if there are some invalid mats
    returns 4 if there are some nans
    returns 5 win!
    """
    identity_mats = ctv_identity_mats(ctv_mm_mult(m, np.transpose(m, (0,2,1))))
    # make sure matrix is self-inverting and the determinate is either 1 in all cases or -1 in all cases
    idx = np.argwhere(ctv_err_test(ctv_determ_mats(m),1) & identity_mats)
    c_right = idx.shape[0]
    idx = np.argwhere(ctv_err_test(ctv_determ_mats(m),-1) & identity_mats)
    c_left = idx.shape[0]
    idx = np.argwhere(~np.isfinite(ctv_determ_mats(m)))
    c_nan = idx.shape[0]
    if (c_left != 0) and (c_right != 0): # mixed system
        return 0
    elif (c_left == 0) and (c_right == 0): # all matrices fail
        return 1
    elif c_nan == m.shape[0]: # all nans
        return 2
    elif (c_left+c_right+c_nan < 0): # some matrices fail
        return 3
    elif c_nan != 0: # some nans
        return 4
    else: # all mats are rotation mats and there is no missing data
        return 5


def ctv_left_mats(m):
    """
    Is this a set of left-handed permutation matrices?
    """
    idx = np.argwhere(ctv_err_test(ctv_determ_mats(m),-1))
    c = idx.shape[0]
    if c > 0:
        return 1
    else:
        return 0


def ctv_swap_hands(m):
    """
    Turns a 3x3 matrix with a left-handed basis into a right-handed basis and vice-versa
    """
    out = m.copy()
    out[:,0,:] *= -1
    return out


def ctv_norm_vec_rot(v):
    """
    Helper function
    Calculates the norm of a bunch of vectors simultaneously
    """
    if v is None:
        return -1
    if v.ndim != 2:
        return -1
    return np.sqrt(np.sum(v**2, axis=1))


def ctv_normalize_vec_rot(v):
    """
    Helper function
    Normalizes a bunch of vectors simultaneously
    """
    if v is None:
        return -1
    if v.ndim != 2:
        return -1
    n_a = ctv_norm_vec_rot(v)
    if (n_a.ndim == 0) and (n_a == -1):
        return -1
    v_s = v.shape
    # calculation is pretty straight forward
    # we turn n_a into an N x D so computation can be done element by element
    n_b = np.repeat(n_a, v_s[1]).reshape(v_s)
    return v/n_b


def ctv_mx_vec_rot(m, x):
    """
    Helper function
    Vectorized fx to multiply n matrices by n vectors
    """
    # input checks
    if m is None:
        return -1
    if x is None:
        return -1
    m_s = m.shape
    x_s = x.shape
    # make sure number of dimensions in input arrays is correct
    if len(m_s) != 3:
        return -1
    if len(x_s) != 2:
        return -1
    # make sure dimensions match
    if not np.array_equal(x_s, [m_s[0], m_s[1]]):
        return -1
    if not np.array_equal(m_s, [x_s[0], x_s[1], x_s[1]]):
        return -1
    # calculation is pretty straight forward
    # we turn x into an N x 3 x 3 so computation can be done element by element
    y_t = np.repeat(x, x_s[1]*x_s[1]).reshape(x_s[0], x_s[1], x_s[1])
    # custom multiplication requires rebin to stack vector across rows,
    # not columns
    y_t = np.transpose(y_t, (0, 2, 1))
    # 9 multiplications and 3 additions per matrix
    y = np.sum(y_t*m, axis=2)
    return y




import unittest
import numpy as np
from pyspedas.cotrans.quaternions import qslerp, qcompose, qconj, mtoq, qtom, qvalidate, qmult, qdecompose


class Qtests(unittest.TestCase):
    def test_qslerp(self):
        m1 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        m2 = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]])
        qin = np.transpose(np.array([mtoq(m1), mtoq(m2)])).reshape([4, 2]).transpose()
        x1 = np.array([0.0, 1.0])
        x2 = np.array([0.0, 1.0/6.0, 1.0/3.0, 1.0/2.0, 2.0/3.0, 5.0/6.0, 1.0])
        qout = qslerp(qin, x1, x2)
        qout_geo = qslerp(qin, x1, x2, geometric=True)
        mout = qtom(qout)
        qcomp = qcompose(np.array([[1, 2, 3]]), np.array([4]), free=False)
        self.assertTrue(np.abs(np.sum(qout-qout_geo))<1e-6)

    def test_errors(self):
        qs1 = np.ones((7, 5))
        qs2 = np.ones((8, 4))
        qs3 = np.ones(4)
        self.assertTrue(qvalidate(qs1, 'qs1', 'qslerp') == -1)
        self.assertTrue(qmult(qs1, qs2) == -1)
        self.assertTrue(qdecompose(np.array(1)) == -1)
        self.assertTrue(qdecompose(1) == -1)
        self.assertTrue(qconj(1) == -1)
        self.assertTrue(qslerp(qs1, qs1, qs1) == -1)


if __name__ == '__main__':
    unittest.main()


import unittest
import numpy as np
from pyspedas.mms import mms_load_state, mms_load_mec, mms_load_fgm, mms_load_scm, mms_load_fpi, mms_load_hpca, mms_load_feeps, mms_load_edp, mms_load_edi, mms_load_aspoc, mms_load_dsp
from pytplot import data_exists
from pyspedas.mms.hpca.mms_hpca_calc_anodes import mms_hpca_calc_anodes
from pyspedas.mms.hpca.mms_hpca_spin_sum import mms_hpca_spin_sum
from pyspedas.mms.hpca.mms_get_hpca_info import mms_get_hpca_info
from pyspedas import tdpwrspc
import pyspedas
from pytplot import get_data, del_data, tplot
import logging


class FSMLoadTestCases(unittest.TestCase):
    def test_load_basic(self):
        files = pyspedas.mms.fsm(trange=['2015-10-16/06:00', '2015-10-16/06:05'], available=True)
        fsm = pyspedas.mms.fsm(trange=['2015-10-16/06:00', '2015-10-16/06:05'])
        self.assertTrue(data_exists('mms1_fsm_b_mag_brst_l3'))
        self.assertTrue(data_exists('mms1_fsm_b_gse_brst_l3'))
        tplot(['mms1_fsm_b_mag_brst_l3', 'mms1_fsm_b_gse_brst_l3'], display=False)


class StateLoadTestCases(unittest.TestCase):
    def test_load_eph_no_update(self):
        data = mms_load_state(datatypes=['pos', 'vel']) # ensure the files are stored locally
        del_data('*') # remove the current tplot vars
        data = mms_load_state(datatypes=['pos', 'vel'], no_update=True) # ensure the files are stored locally
        self.assertTrue(data_exists('mms1_defeph_pos'))
        self.assertTrue(data_exists('mms1_defeph_vel'))

    def test_load_eph_data(self):
        data = mms_load_state(datatypes=['pos', 'vel'])
        self.assertTrue(data_exists('mms1_defeph_pos'))
        self.assertTrue(data_exists('mms1_defeph_vel'))
        tplot(['mms1_defeph_pos', 'mms1_defeph_vel'], display=False)

    def test_load_att_data(self):
        data = mms_load_state(trange=['2015-10-16', '2015-10-16/06:00'], datatypes=['spinras', 'spindec'])
        self.assertTrue(data_exists('mms1_defatt_spinras'))
        self.assertTrue(data_exists('mms1_defatt_spindec'))
        tplot(['mms1_defatt_spinras', 'mms1_defatt_spindec'], display=False)


############### DSP ############### 
class DSPLoadTestCases(unittest.TestCase):
    # def test_load_default_data(self):
    #     data = mms_load_dsp(trange=['2015-08-01','2015-09-01'], datatype='swd', level='l2', data_rate='fast')
    #     self.assertTrue(data_exists('mms1_dsp_swd_E12_Counts'))

    def test_load_epsd_bpsd_data(self):
        data = mms_load_dsp(trange=['2015-08-01','2015-08-02'], datatype=['epsd', 'bpsd'], level='l2', data_rate='fast')
        self.assertTrue(data_exists('mms1_dsp_epsd_omni'))
        self.assertTrue(data_exists('mms1_dsp_bpsd_omni'))
        tplot(['mms1_dsp_epsd_omni', 'mms1_dsp_bpsd_omni'], display=False)

    def test_load_bpsd_data(self):
        data = mms_load_dsp(trange=['2015-10-16','2015-10-17'], datatype='bpsd', level='l2', data_rate='fast', available=True)
        data = mms_load_dsp(trange=['2015-10-16','2015-10-17'], datatype='bpsd', level='l2', data_rate='fast')
        self.assertTrue(data_exists('mms1_dsp_bpsd_omni_fast_l2'))
        tplot(['mms1_dsp_bpsd_omni_fast_l2'], display=False)

    def test_load_epsd_spdf(self):
        data = mms_load_dsp(trange=['2015-08-01','2015-08-02'], datatype='epsd', level='l2', data_rate='fast', spdf=True)
        self.assertTrue(data_exists('mms1_dsp_epsd_omni'))
        tplot(['mms1_dsp_epsd_omni'], display=False)

    def test_load_epsd_suffix(self):
        data = mms_load_dsp(trange=['2015-08-01','2015-08-02'], datatype='epsd', level='l2', data_rate='fast', suffix='_test')
        self.assertTrue(data_exists('mms1_dsp_epsd_omni_test'))
        tplot(['mms1_dsp_epsd_omni_test'], display=False)


############### FEEPS ############### 
# class FEEPSLoadTestCases(unittest.TestCase):
#     def test_load_default_data(self):
#         data = mms_load_feeps(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
#         data = mms_load_feeps(trange=['2015-10-16', '2015-10-16/01:00'])
#         self.assertTrue(data_exists('mms1_epd_feeps_srvy_l2_electron_intensity_omni'))
#         self.assertTrue(data_exists('mms1_epd_feeps_srvy_l2_electron_intensity_omni_spin'))
#
#     def test_load_spdf_data(self):
#         del_data('*')
#         data = mms_load_feeps(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', spdf=True)
#         self.assertTrue(data_exists('mms1_epd_feeps_brst_l2_electron_intensity_omni'))
#         self.assertTrue(data_exists('mms1_epd_feeps_brst_l2_electron_intensity_omni_spin'))
#
#     def test_load_suffix(self):
#         data = mms_load_feeps(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', suffix='_test')
#         self.assertTrue(data_exists('mms1_epd_feeps_brst_l2_electron_intensity_omni_test'))
#         self.assertTrue(data_exists('mms1_epd_feeps_brst_l2_electron_intensity_omni_spin_test'))
#
#     def test_load_brst_ion_data(self):
#         del_data('*')
#         data = mms_load_feeps(probe=4, data_rate='brst', datatype='ion', trange=['2015-10-01/10:48:16', '2015-10-01/10:49:16'])
#         self.assertTrue(data_exists('mms4_epd_feeps_brst_l2_ion_intensity_omni'))
#         self.assertTrue(data_exists('mms4_epd_feeps_brst_l2_ion_intensity_omni_spin'))
#
#     def test_load_brst_multi_probe(self):
#         del_data('*')
#         data = mms_load_feeps(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', probe=[1, 2, 3, 4])
#         self.assertTrue(data_exists('mms1_epd_feeps_brst_l2_electron_intensity_omni'))
#         self.assertTrue(data_exists('mms1_epd_feeps_brst_l2_electron_intensity_omni_spin'))
#         self.assertTrue(data_exists('mms2_epd_feeps_brst_l2_electron_intensity_omni'))
#         self.assertTrue(data_exists('mms2_epd_feeps_brst_l2_electron_intensity_omni_spin'))
#         self.assertTrue(data_exists('mms3_epd_feeps_brst_l2_electron_intensity_omni'))
#         self.assertTrue(data_exists('mms3_epd_feeps_brst_l2_electron_intensity_omni_spin'))
#         self.assertTrue(data_exists('mms4_epd_feeps_brst_l2_electron_intensity_omni'))
#         self.assertTrue(data_exists('mms4_epd_feeps_brst_l2_electron_intensity_omni_spin'))



############### HPCA ############### 
class HPCALoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_hpca(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_hpca(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_hpca_hplus_number_density'))
        tplot(['mms1_hpca_hplus_number_density'], display=False)

    def test_load_spdf_data(self):
        data = mms_load_hpca(trange=['2015-10-16', '2015-10-16/01:00'], spdf=True)
        self.assertTrue(data_exists('mms1_hpca_hplus_number_density'))
        tplot(['mms1_hpca_hplus_number_density'], display=False)

    def test_load_ion_omni_suffix(self):
        del_data('*')
        data = mms_load_hpca(probe=2, trange=['2016-08-09/09:10', '2016-08-09/10:10:00'], datatype='ion', data_rate='brst', suffix='_brst')
        mms_hpca_calc_anodes(fov=[0, 360], probe=2, suffix='_brst')
        mms_hpca_spin_sum(probe=2, suffix='_brst', avg=True)
        self.assertTrue(data_exists('mms2_hpca_hplus_flux_brst_elev_0-360_spin'))
        tplot(['mms2_hpca_hplus_flux_brst_elev_0-360_spin'], display=False)

    def test_load_ion_omni(self):
        del_data('*')
        data = mms_load_hpca(trange=['2016-10-16', '2016-10-16/6:00'], datatype='ion')
        mms_hpca_calc_anodes(fov=[0, 360], probe='1')
        mms_hpca_spin_sum()
        self.assertTrue(data_exists('mms1_hpca_hplus_flux_elev_0-360_spin'))
        self.assertTrue(data_exists('mms1_hpca_heplus_flux_elev_0-360_spin'))
        self.assertTrue(data_exists('mms1_hpca_heplusplus_flux_elev_0-360_spin'))
        self.assertTrue(data_exists('mms1_hpca_oplus_flux_elev_0-360_spin'))
        tplot(['mms1_hpca_hplus_flux_elev_0-360_spin',
               'mms1_hpca_heplus_flux_elev_0-360_spin',
               'mms1_hpca_heplusplus_flux_elev_0-360_spin',
               'mms1_hpca_oplus_flux_elev_0-360_spin'], display=False)

    def test_center_fast_moments_data(self):
        data = mms_load_hpca(trange=['2015-10-16/14:00', '2015-10-16/15:00'])
        centered = mms_load_hpca(trange=['2015-10-16/14:00', '2015-10-16/15:00'], center_measurement=True, suffix='_centered')
        
        t, d = get_data('mms1_hpca_hplus_ion_bulk_velocity')
        c, d = get_data('mms1_hpca_hplus_ion_bulk_velocity_centered')
        self.assertTrue(np.round(c[0]-t[0], decimals=3) == 5.0)

    def test_center_brst_moments_data(self):
        data = mms_load_hpca(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst')
        centered = mms_load_hpca(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', center_measurement=True, suffix='_centered')
        
        t, d = get_data('mms1_hpca_hplus_ion_bulk_velocity')
        c, d = get_data('mms1_hpca_hplus_ion_bulk_velocity_centered')
        self.assertTrue(np.round(c[0]-t[0], decimals=3) == 5.0)

    def test_hpca_info(self):
        info = mms_get_hpca_info()
        self.assertTrue(list(info.keys()) == ['elevation', 't_spin', 't_sweep', 'azimuth_energy_offset'])


############### EDI ############### 
class EDILoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'], available=True)
        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'])
        self.assertTrue(data_exists('mms1_edi_e_gse_srvy_l2'))
        tplot(['mms1_edi_e_gse_srvy_l2'], display=False)

    def test_load_spdf_data(self):
        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'], spdf=True)
        self.assertTrue(data_exists('mms1_edi_e_gse_srvy_l2'))
        tplot(['mms1_edi_e_gse_srvy_l2'], display=False)

    def test_load_suffix(self):
        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_edi_e_gse_srvy_l2_test'))
        tplot(['mms1_edi_e_gse_srvy_l2_test'], display=False)


############### ASPOC ############### 
class ASPOCLoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_aspoc_ionc_l2'))
        tplot(['mms1_aspoc_ionc_l2'], display=False)

    def test_load_spdf_data(self):
        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'], spdf=True)
        self.assertTrue(data_exists('mms1_aspoc_ionc_l2'))

    def test_load_suffix(self):
        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_aspoc_ionc_l2_test'))


############### EDP ############### 
class EDPLoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))
        tplot(['mms1_edp_dce_gse_fast_l2'], display=False)

    def test_load_hfesp_data(self):
        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'], datatype='hfesp', data_rate='srvy')
        self.assertTrue(data_exists('mms1_edp_hfesp_srvy_l2'))
        tplot(['mms1_edp_hfesp_srvy_l2'], display=False)

    def test_load_spdf_data(self):
        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'], spdf=True)
        self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))
        tplot(['mms1_edp_dce_gse_fast_l2'], display=False)

    def test_load_suffix(self):
        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))

    # def test_load_multiple_sc(self):
    #     data = mms_load_edp(probe=['1', '2', '3', '4'], trange=['2015-10-16', '2015-10-16/01:00'])
    #     self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))
    #     self.assertTrue(data_exists('mms2_edp_dce_gse_fast_l2'))
    #     self.assertTrue(data_exists('mms3_edp_dce_gse_fast_l2'))
    #     self.assertTrue(data_exists('mms4_edp_dce_gse_fast_l2'))

    def test_load_brst_data(self):
        data = mms_load_edp(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'])
        self.assertTrue(data_exists('mms1_edp_dce_gse_brst_l2'))
        tplot(['mms1_edp_dce_gse_brst_l2'], display=False)


############### FGM ############### 
class FGMLoadTestCases(unittest.TestCase):
    def test_regression_multi_imports_spdf(self):
        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], spdf=True)
        t1, d1 = get_data('mms1_fgm_b_gse_brst_l2')
        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], spdf=True)
        t2, d2 = get_data('mms1_fgm_b_gse_brst_l2')
        self.assertTrue(t1.shape == t2.shape)
        self.assertTrue(d1.shape == d2.shape)

    def test_load_default_data(self):
        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'],available=True)
        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_fgm_b_gse_srvy_l2'))
        self.assertTrue(data_exists('Epoch'))
        self.assertTrue(data_exists('Epoch_state'))
        tplot(['mms1_fgm_b_gse_srvy_l2'], display=False)

    def test_load_default_data_exclude(self):
        # Capture all log messages of level INFO or above
        with self.assertLogs(level=logging.INFO) as captured:
            # assertLogs fails if there are no log messages, so we make sure there's at least one
            logging.info("Dummy log message")
            data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'],exclude_format='*rdeltahalf*',available=True)
            data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'],exclude_format='*rdeltahalf*')
            self.assertTrue(data_exists('mms1_fgm_b_gse_srvy_l2'))
            self.assertTrue(data_exists('Epoch'))
            self.assertTrue(data_exists('Epoch_state'))
            tplot(['mms1_fgm_b_gse_srvy_l2'], display=False)
        # Assert that none of the log messages contain the string "rdeltahalf"
        logging.info("Captured log messages:")
        for rec in captured.records:
            logging.info(rec.msg)
            self.assertTrue("rdeltahalf" not in rec.msg)

    def test_load_spdf_data(self):
        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], spdf=True)
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2'))

    def test_load_suffix(self):
        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], suffix='_test')
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2_test'))

    def test_load_multiple_sc(self):
        data = mms_load_fgm(probe=[1, 2, 3, 4], data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'])
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2'))
        self.assertTrue(data_exists('mms2_fgm_b_gse_brst_l2'))
        self.assertTrue(data_exists('mms3_fgm_b_gse_brst_l2'))
        self.assertTrue(data_exists('mms4_fgm_b_gse_brst_l2'))

    def test_load_brst_data(self):
        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'])
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2'))
        tplot(['mms1_fgm_b_gse_brst_l2'], display=False)

    def test_load_data_no_update(self):
        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00']) # make sure the files exist locally
        del_data('*') 
        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'], no_update=True) # load the file from the local cache
        self.assertTrue(data_exists('mms1_fgm_b_gse_srvy_l2'))


############### MEC ###############
class MECLoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_mec(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_mec(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_mec_r_sm'))
        tplot(['mms1_mec_r_sm'], display=False)

    def test_load_spdf_data(self):
        data = mms_load_mec(trange=['2015-10-16', '2015-10-16/01:00'], spdf=True)
        self.assertTrue(data_exists('mms1_mec_r_sm'))

    def test_load_suffix(self):
        data = mms_load_mec(trange=['2015-10-16', '2015-10-16/01:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_mec_r_sm_test'))


class SCMLoadTestCases(unittest.TestCase):
    def test_brst_dpwrspc_data(self):
        data = mms_load_scm(probe=4, data_rate='brst', datatype='scb', trange=['2015-10-01/10:48:16', '2015-10-01/10:49:16'])
        tdpwrspc('mms4_scm_acb_gse_scb_brst_l2', notmvariance=True)
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_x_dpwrspc'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_y_dpwrspc'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_z_dpwrspc'))
        tplot(['mms4_scm_acb_gse_scb_brst_l2',
               'mms4_scm_acb_gse_scb_brst_l2_x_dpwrspc',
               'mms4_scm_acb_gse_scb_brst_l2_y_dpwrspc',
               'mms4_scm_acb_gse_scb_brst_l2_z_dpwrspc'], display=False)

    def test_load_default_data(self):
        data = mms_load_scm(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_scm(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_scm_acb_gse_scsrvy_srvy_l2'))

    def test_load_schb(self):
        data = pyspedas.mms.scm(probe=4, data_rate='brst', datatype='schb', trange=['2015-10-01/10:48:16', '2015-10-01/10:49:16'])
        self.assertTrue(data_exists('mms4_scm_acb_gse_schb_brst_l2'))

    def test_load_suffix(self):
        data = mms_load_scm(trange=['2015-10-16', '2015-10-16/01:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_scm_acb_gse_scsrvy_srvy_l2_test'))

    def test_load_multiple_sc(self):
        data = mms_load_scm(probe=['1', '2', '3', '4'], trange=['2017-12-15', '2017-12-16'])
        # self.assertTrue(data_exists('mms1_scm_acb_gse_scsrvy_srvy_l2'))
        # self.assertTrue(data_exists('mms2_scm_acb_gse_scsrvy_srvy_l2'))
        self.assertTrue(data_exists('mms3_scm_acb_gse_scsrvy_srvy_l2'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scsrvy_srvy_l2'))

    def test_load_brst_data(self):
        data = mms_load_scm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], datatype='scb')
        self.assertTrue(data_exists('mms1_scm_acb_gse_scb_brst_l2'))

    def test_available(self):
        files = mms_load_scm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], datatype='scb', available=True)
        self.assertTrue(len(files) == 2)


if __name__ == '__main__':
    unittest.main()



import logging
import pytplot

def tnormalize(variable, newname=None, return_data=False):
    """
    Normalize all the vectors stored in a tplot variable

    Input
    ----------
        variable: str or np.ndarray
            tplot variable (or numpy array) containing the vectors to be normalized

    Parameters
    ----------
        newname: str
            name of the output variable; default: variable_normalized

        return_data: bool
            return the normalized vectors instead of creating a tplot variable
    
    Returns
    ----------

        name of the tplot variable created or normalized vectors if return_data
        is set

    """
    logging.info("tnormalize has been moved to the pytplot.tplot_math module. Please update your imports!")
    logging.info("This version will eventually be removed.")
    return pytplot.tplot_math.tnormalize(variable=variable,newname=newname,return_data=return_data)


import logging
import numpy as np
from scipy.interpolate import NearestNDInterpolator
from astropy.coordinates import spherical_to_cartesian


def spd_pgs_regrid(data, regrid_dimen):
    """

    """
    if len(regrid_dimen) != 2:
        logging.error('Invalid regrid dimensions; the dimensions should be [n_phi, n_theta]')
        return

    n_energy = len(data['energy'][:, 0])

    n_phi_grid = int(regrid_dimen[0])
    n_theta_grid = int(regrid_dimen[1])
    n_bins_grid = n_phi_grid*n_theta_grid

    d_phi_grid = 360.0/n_phi_grid
    d_theta_grid = 180.0/n_theta_grid

    phi_angles = (np.arange(n_bins_grid) % n_phi_grid)*d_phi_grid+d_phi_grid/2.0
    phi_grid = np.repeat(np.reshape(phi_angles, [n_bins_grid, 1]), n_energy, axis=1).T

    theta_angles = np.fix(np.arange(n_bins_grid)/n_phi_grid)*d_theta_grid+d_theta_grid/2.0 - 90
    theta_grid = np.repeat(np.reshape(theta_angles, [n_bins_grid, 1]), n_energy, axis=1).T

    d_phi_grid = np.zeros([n_energy, n_bins_grid]) + d_phi_grid
    d_theta_grid = np.zeros([n_energy, n_bins_grid]) + d_phi_grid

    data_grid = np.zeros([n_energy, n_bins_grid])
    bins_grid = np.zeros([n_energy, n_bins_grid])

    output = {'data': data_grid,
              'scaling': data_grid,
              'phi': phi_grid,
              'dphi': d_phi_grid,
              'theta': theta_grid,
              'dtheta': d_theta_grid,
              'energy': data_grid,
              'denergy': data_grid,
              'bins': bins_grid}

    # assumes energies are constant across angle
    output['energy'] = np.repeat(np.reshape(data['energy'][:, 0], [n_energy, 1]), n_bins_grid, axis=1).T
    output['denergy'] = np.repeat(np.reshape(data['denergy'][:, 0], [n_energy, 1]), n_bins_grid, axis=1).T

    for i in range(0, n_energy):
        phi_temp = data['phi'][i, :]
        theta_temp = data['theta'][i, :]
        data_temp = data['data'][i, :]
        bins_temp = data['bins'][i, :]

        r_grid = np.ones(len(phi_grid[i, :]))

        data_grid_interp = griddata(phi_temp, theta_temp, data_temp)
        bins_grid_interp = griddata(phi_temp, theta_temp, bins_temp)

        grid_x, grid_y, grid_z = spherical_to_cartesian(r_grid, theta_grid[i, :]*np.pi/180.0, phi_grid[i, :]*np.pi/180.0)

        for j in range(0, len(phi_grid[i, :])):
            output['data'][i, j] = data_grid_interp(grid_x[j], grid_y[j], grid_z[j])
            output['bins'][i, j] = bins_grid_interp(grid_x[j], grid_y[j], grid_z[j])

    return output


def griddata(phi, theta, data):
    r = np.ones(len(phi))
    phi_rad = phi*np.pi/180.0
    theta_rad = theta*np.pi/180.0
    cart_temp = spherical_to_cartesian(r, theta_rad, phi_rad)

    points = np.stack(cart_temp).T
    return NearestNDInterpolator(points, data)



from pyspedas import time_double
from pyspedas.cotrans.minvar import minvar
from pytplot import get_data, store_data
import numpy as np


def minvar_matrix_make(in_var_name,
                       tstart=None,
                       tstop=None,
                       twindow=None,
                       tslide=None,
                       newname=None):
    """

    """
    data = get_data(in_var_name)

    if tstart is None:
        start_d = data.times[0]
    else:
        start_d = time_double(tstart)

    if tstop is None:
        stop_d = data.times[-1]
    else:
        stop_d = time_double(tstop)

    if twindow is None:
        twindow = stop_d - start_d

    # sets it to something large enough that it will only generate one
    # matrix in the default case
    if tslide is None:
        tslide = twindow / 2.0

    if newname is None:
        newname = in_var_name + '_mva_mat'

    current_d = start_d

    # estimate the number of output matrices to generate temporary storage
    if tslide != 0:
        o_num = (stop_d - start_d) / tslide
    else:
        o_num = 1

    o_num = int(o_num)

    o_times = np.zeros(o_num)
    o_lams = np.zeros((o_num, 3))
    o_eigs = np.zeros((o_num, 3, 3))

    i = 0

    while current_d + twindow <= stop_d + twindow:
        if i >= o_num:
            break
        # output time for the mva matrix is the midpoint time for the interval
        o_times[i] = current_d + twindow / 2.0

        idx = np.argwhere((data.times >= current_d) & (data.times <= (current_d + twindow))).flatten()

        if len(idx) == 0:
            current_d += tslide
            i += 1
            continue

        in_data = data.y[idx, :]

        try:
            out = minvar(in_data)
        except np.linalg.LinAlgError:
            current_d += tslide
            i += 1
            continue

        if out is not None:
            v_rot, o_eig, o_lam = out
            o_lams[i, :] = o_lam
            o_eigs[i, :, :] = o_eig.transpose()

        i += 1
        current_d += tslide

        if tslide == 0:
            break

    o_d = {'x': o_times[:-1], 'y': o_eigs[0:-1, :, :]}

    out_vars = []

    saved = store_data(newname, data=o_d)

    if saved:
        out_vars.append(newname)

    return out_vars



from .load import load

def mrmi(trange=['2020-11-5', '2020-11-6'],
         probe='a',
         datatype='mrmi',
         level='l1',
         suffix='',
         get_support_data=False,
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the Magneto Resistive Magnetometer (MRMi)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str
            Spacecraft identifier ('a' or 'b')

        datatype: str
            Data type; Valid options:
                'mrmi' for L1 data

        level: str
            Data level; options: 'l1' (default: l1)

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='mrmi', probe=probe, trange=trange, level=level, datatype=datatype, suffix=suffix,
                 get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly,
                 notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return mrmi_postprocessing(tvars)

def mrmi_postprocessing(variables):
    """
    Placeholder for MRMa post-processing
    """
    return variables



import numpy as np
from pyspedas import time_double


def mms_feeps_active_eyes(trange, probe, data_rate, species, level):
    """
    This function returns the FEEPS active eyes, based on date/probe/species/rate
    
    Parameters:
        trange: list of str
            time range

        probe: str
            probe #, e.g., '4' for MMS4

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst'

        species: str
            'electron' or 'ion'

        level: str
            data level

    Returns:
        Hash table containing 2 keys:
            output['top'] -> maps to the active top eyes
            output['bottom'] -> maps to the active bottom eyes

    Notes:
         1) Burst mode should include all sensors (TOP and BOTTOM):
             electrons: [1, 2, 3, 4, 5, 9, 10, 11, 12]
             ions: [6, 7, 8]
         
         2) SITL should return (TOP only):
             electrons: set_intersection([5, 11, 12], active_eyes)
             ions: None
             
         3) From Drew Turner, 9/7/2017, srvy mode:
         
           - before 16 August 2017:
              electrons: [3, 4, 5, 11, 12]
              iond: [6, 7, 8]
         
           - after 16 August 2017:
               MMS1
                 Top Eyes: 3, 5, 6, 7, 8, 9, 10, 12
                 Bot Eyes: 2, 4, 5, 6, 7, 8, 9, 10

               MMS2
                 Top Eyes: 1, 2, 3, 5, 6, 8, 10, 11
                 Bot Eyes: 1, 4, 5, 6, 7, 8, 9, 11

               MMS3
                 Top Eyes: 3, 5, 6, 7, 8, 9, 10, 12
                 Bot Eyes: 1, 2, 3, 6, 7, 8, 9, 10

               MMS4
                 Top Eyes: 3, 4, 5, 6, 8, 9, 10, 11
                 Bot Eyes: 3, 5, 6, 7, 8, 9, 10, 12

    """
    if data_rate.lower() == 'brst' and species.lower() == 'electron': return {'top': [1, 2, 3, 4, 5, 9, 10, 11, 12], 'bottom': [1, 2, 3, 4, 5, 9, 10, 11, 12]}
    if data_rate.lower() == 'brst' and species.lower() == 'ion': return {'top': [6, 7, 8], 'bottom': [6, 7, 8]}

    # old eyes, srvy mode, prior to 16 August 2017
    if species.lower() == 'electron':
        sensors = {'top': [3, 4, 5, 11, 12], 'bottom': [3, 4, 5, 11, 12]}
    else:
        sensors = {'top': [6, 7, 8], 'bottom': [6, 7, 8]}

    if isinstance(trange[0], str): 
        start_time = time_double(trange[0])
    elif isinstance(trange[0], np.datetime64):
        start_time = np.int64(trange[0]) / 1e9
    else:
        start_time = trange[0]

    # srvy mode, after 16 August 2017
    if start_time >= time_double('2017-08-16') and data_rate.lower() == 'srvy':
        active_table = {'1-electron': {'top': [3, 5, 9, 10, 12], 'bottom': [2, 4, 5, 9, 10]}, 
        '1-ion': {'top': [6, 7, 8], 'bottom': [6, 7, 8]},
        '2-electron': {'top': [1, 2, 3, 5, 10, 11], 'bottom': [1, 4, 5, 9, 11]},
        '2-ion': {'top': [6, 8], 'bottom': [6, 7, 8]},
        '3-electron': {'top': [3, 5, 9, 10, 12], 'bottom': [1, 2, 3, 9, 10]},
        '3-ion': {'top': [6, 7, 8], 'bottom': [6, 7, 8]},
        '4-electron': {'top': [3, 4, 5, 9, 10, 11], 'bottom': [3, 5, 9, 10, 12]},
        '4-ion': {'top': [6, 8], 'bottom': [6, 7, 8]}}
        sensors = active_table[probe.lower()+'-'+species.lower()]
        if level.lower() == 'sitl':
            return {'top': list(set(sensors['top']) & set([5, 11, 12])), 'bottom': []}

    if level.lower() == 'sitl':
        return {'top': [5, 11, 12], 'bottom': []}

    return sensors



import unittest
from pytplot import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_mag_data(self):
        vfm_vars = pyspedas.swarm.mag(probe='c',
                                      trange=['2017-03-27/06:00', '2017-03-27/08:00'],
                                      datatype='hr',
                                      time_clip=True)
        self.assertTrue(data_exists('swarmc_B_VFM'))


if __name__ == '__main__':
    unittest.main()




from .load import load
from .filter import filter_fields

from pytplot import options, data_quants

# Loading
def fields(trange=['2018-11-5', '2018-11-6'], 
        datatype='mag_rtn', 
        level='l2',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        username=None,
        password=None,
        last_version=False
        ):
    """
    This function loads Parker Solar Probe FIELDS data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options include:
                'mag_RTN'
                'mag_RTN_1min'
                'mag_rtn_4_per_cycle' (SPDF only)
                'mag_RTN_4_Sa_per_Cyc' 
                'mag_SC'
                'mag_SC_1min'
                'mag_sc_4_per_cycle' (SPDF only)
                'mag_SC_4_Sa_per_Cyc' 
                'mag_VSO' (limited dates)
                'rfs_burst' (limited dates)
                'rfs_hfr', 
                'rfs_lfr'
                'f2_100bps'
                'dfb_dc_spec'
                'dfb_ac_spec'
                'dfb_dc_xspec'
                'dfb_ac_xspec'
                'merged_scam_wf'
                'sqtn_rfs_V1V2'

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, this flag is False but 
            FIELDS support data is always loaded for datatypes where filtering
            on quality flags is supported.

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        username: str
            Username to use for authentication.

            If passed, attempt to download data from the FIELDS Instrument Team server
            instead of the fully public server at SPDF.
            Provides access to unpublished, V02 files.

            Implemented for dataypes:
                'mag_RTN_1min'
                'mag_RTN_4_Sa_per_Cyc'
                'mag_SC'
                'mag_SC_1min'
                'mag_SC_4_Sa_per_Cyc'
                'sqtn_rfs_V1V2'

        password: str
            Password to use for authentication

        last_version: bool
            If True, only download the highest-numbered file version
    Returns
    ----------
        List of tplot variables created.

    """

    if suffix:
        suffix = '_' + suffix

    # SCaM and QTN data are Level 3
    if datatype.lower() in ['merged_scam_wf', 'sqtn_rfs_v1v2']:
        level = 'l3'
        print("Using LEVEL=L3")

    spec_types = None
    if datatype in ['dfb_dc_spec', 'dfb_ac_spec', 'dfb_dc_xspec', 'dfb_ac_xspec']:
        if level == 'l1':
            spec_types = ['1', '2', '3', '4']
        else:
            if datatype == 'dfb_dc_spec' or datatype == 'dfb_ac_spec':
                spec_types = ['dV12hg','dV34hg','dV12lg','dV34lg',
                    'SCMulfhg','SCMvlfhg','SCMwlfhg',
                    'SCMulflg','SCMvlflg','SCMwlflg',
                    'SCMdlfhg','SCMelfhg','SCMflfhg',
                    'SCMdlflg','SCMelflg','SCMflflg',
                    'SCMmf', 'V5hg']
            else:
                spec_types = ['SCMdlfhg_SCMelfhg','SCMdlfhg_SCMflfhg','SCMelfhg_SCMflfhg',
                    'SCMulfhg_SCMvlfhg','SCMulfhg_SCMwlfhg','SCMvlfhg_SCMwlfhg',
                    'dV12hg_dV34hg']

    loaded_vars = load(
        instrument='fields', trange=trange, datatype=datatype, spec_types=spec_types, level=level, 
        suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, 
        downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update,
        username=username, password=password,last_version=last_version
    )
    
    if loaded_vars is None or notplot or downloadonly:
        return loaded_vars

    qf_root = 'psp_fld_l2_quality_flags'+suffix

    # If variables are loaded that quality flag filtering supports --
    # Make sure the quality flag variable is also loaded and linked. 
    mag_rtnvars = [x for x in loaded_vars if 'fld_l2_mag_RTN' in x ]
    mag_scvars = [x for x in loaded_vars if 'fld_l2_mag_SC' in x ]
    rfs_vars = [x for x in loaded_vars if 'rfs_lfr' in x or 'rfs_hfr' in x]
    if (len(mag_rtnvars + mag_scvars + rfs_vars) > 0) \
        & ('psp_fld_l2_quality_flags'+suffix not in loaded_vars):
        loaded_extra = load(
            instrument='fields', trange=trange, datatype=datatype, spec_types=spec_types, level=level, 
            suffix=suffix, get_support_data=True, varformat=varformat, varnames=['psp_fld_l2_quality_flags'],
            downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update,
            username=username, password=password,last_version=last_version
        )
        qf_root = 'psp_fld_l2_quality_flags'+suffix if 'psp_fld_l2_quality_flags'+suffix in loaded_extra else None
        loaded_vars += loaded_extra

    for var in mag_rtnvars:
        options(var, 'legend_names', ['Br (RTN)', 'Bt (RTN)', 'Bn (RTN)'])
        data_quants[var] = data_quants[var].assign_attrs({'qf_root':qf_root})

    for var in mag_scvars:
        options(var, 'legend_names', ['Bx', 'By', 'Bz'])
        data_quants[var] = data_quants[var].assign_attrs({'qf_root':qf_root})

    for var in rfs_vars:
        data_quants[var] = data_quants[var].assign_attrs({'qf_root':qf_root})

    return loaded_vars


def spc(trange=['2018-11-5', '2018-11-6'], 
        datatype='l3i', 
        level='l3',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        username=None,
        password=None,
        last_version=False
    ):
    """
    This function loads Parker Solar Probe Solar Probe Cup data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options include:
                'l3i' (level='l3')
                'l2i' (level='l2')

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        username: str
            Username to use for authentication.
            
            If passed, attempt to download data from the SWEAP Instrument Team server
            instead of the fully public server at SPDF.
            Provides access to unpublished files.

        password: str
            Password to use for authentication

        last_version: bool
            If True, only download the highest-numbered file version

    Returns
    ----------
        List of tplot variables created.

    """

    if datatype == 'l3i':
        level = 'l3'
        print("Using LEVEL=L3")
    elif datatype == 'l2i':
        level = 'l2'
        print("Using LEVEL=L2")

    return load(instrument='spc', trange=trange, datatype=datatype, level=level, suffix=suffix, 
        get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, 
        notplot=notplot, time_clip=time_clip, no_update=no_update, username=username, password=password, last_version=last_version)

def spe(trange=['2018-11-5', '2018-11-6'], 
        datatype='spa_sf1_32e', 
        level='l2',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        last_version=False
    ):
    """
    This function loads Parker Solar Probe SWEAP/SPAN-e data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options include:
                spa_sf0_pad  (L3)
                spb_sf0_pad  (L3)
                spe_sf0_pad  (L3)
                spa_sf1_32e  (L2)
                spb_sf1_32e  (L2)
                spa_sf0_16ax8dx32e  (L2)
                spb_sf0_16ax8dx32e  (L2)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        last_version: bool
            If True, only download the highest-numbered file version

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='spe', trange=trange, datatype=datatype, level=level, 
        suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, 
        downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update,last_version=last_version)

def spi(trange=['2018-11-5', '2018-11-6'], 
        datatype='sf00_l3_mom', 
        level='l3',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        username=None,
        password=None,
        last_version=False
    ):
    """
    This function loads Parker Solar Probe SWEAP/SPAN-i data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options Include:
                'sf00_l3_mom': Moments of the Proton distribution function (RTN)
                'sf0a_l3_mom': Moments of the Alpha distribution function (RTN)
                'sf00_l3_mom_inst': Moments of the Proton distribution function (Instrument Frame)
                'sf0a_l3_mom_inst': Moments of the Alpha distribution function (Instrument Frame)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        username: str
            Username to use for authentication.
            
            If passed, attempt to download data from the SWEAP Instrument Team server
            instead of the fully public server at SPDF.
            Provides access to unpublished files.

        password: str
            Password to use for authentication

        last_version: bool
            If True, only download the highest-numbered file version

    Returns
    ----------
        List of tplot variables created.

    """
    if datatype in ['sf00_l3_mom','sf0a_l3_mom','sf00_l3_mom_inst','sf0a_l3_mom_inst']:
        datatype = 'spi_' + datatype
        level = 'l3'
        print("Using LEVEL=L3")

    return load(instrument='spi', trange=trange, datatype=datatype, level=level, 
        suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, 
        downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, 
        username=username, password=password,last_version=last_version
        )

def epihi(trange=['2018-11-5', '2018-11-6'], 
        datatype='let1_rates1h', 
        level='l2',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        last_version=False):
    """
    This function loads Parker Solar Probe ISoIS/EPI-Hi data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        last_version: bool
            If True, only download the highest-numbered file version

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='epihi', trange=trange, datatype=datatype, level=level, 
        suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, 
        downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update,last_version=last_version)

def epilo(trange=['2018-11-5', '2018-11-6'], 
        datatype='pe', 
        level='l2',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        last_version=False):
    """
    This function loads Parker Solar Probe ISoIS/EPI-Lo data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        last_version: bool
            If True, only download the highest-numbered file version

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='epilo', trange=trange, datatype=datatype, level=level, 
        suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, 
        downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update,last_version=last_version)

def epi(trange=['2018-11-5', '2018-11-6'], 
        datatype='summary', 
        level='l2',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        last_version=False):
    """
    This function loads Parker Solar Probe ISoIS/EPI (merged summary) data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        last_version: bool
            If True, only download the highest-numbered file version

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='epi', trange=trange, datatype=datatype, level=level, 
        suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, 
        downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update,last_version=last_version)

